from collections import OrderedDict

from PyQt4 import QtCore, QtGui
from Orange.widgets import widget, gui

from skfusion import fusion
from orangecontrib.datafusion.models import Relation, RelationCompleter
from orangecontrib.datafusion.widgets.owfusiongraph import \
    relation_str

import numpy as np


def _rmse(true, pred):
    return np.sqrt(np.sum((true - pred)**2) / true.size)


def RMSE(rel_true, rel_pred):
    """ NaN-skipping RMSE of masked values between the original relation `rel_true`
        and fuser-completed relation `rel_pred`.
    """
    assert np.ma.is_masked(rel_true)
    test_idx = np.logical_and(rel_true.mask, ~np.isnan(rel_true))
    test = rel_true.data[test_idx]
    pred = rel_pred[test_idx]
    mn, mx = np.min(test), np.max(test)
    pred = np.clip(pred, mn, mx)
    return _rmse(test, pred)


class OWCompletionScoring(widget.OWWidget):
    name = 'Completion Scoring'
    description = "Score the quality of matrix completion using " \
                  "root mean squared error (RMSE)."
    priority = 40000
    icon = 'icons/CompletionScoring.svg'
    inputs = [
        ('Fitted fusion graph', RelationCompleter, 'on_fuser_change', widget.Multiple),
        ('Relation', Relation, 'on_relation_change', widget.Multiple),
    ]

    want_main_area = True
    want_control_area = False

    def __init__(self):
        super().__init__()
        self.fusers = OrderedDict()
        self.relations = OrderedDict()
        self._create_layout()

    def _create_layout(self):
        box = gui.widgetBox(self.mainArea, 'RMSE')
        BOLD_FONT = QtGui.QFont()
        BOLD_FONT.setWeight(QtGui.QFont.DemiBold)
        widget = self

        class HereTableWidget(QtGui.QTableWidget):
            def __init__(self, parent):
                super().__init__(parent)
                parent.layout().addWidget(self)
                self.setHorizontalScrollMode(self.ScrollPerPixel)
                self.setVerticalScrollMode(self.ScrollPerPixel)

            def update_table(self, fusers, relations):
                self.clear()
                self.setRowCount(0)
                self.setColumnCount(len(fusers))
                self.setHorizontalHeaderLabels([fuser[0].name for fuser in fusers.values()])
                for id, relation in relations.items():
                    row = self.rowCount()
                    self.insertRow(row)
                    if not np.ma.is_masked(relation.data):
                        widget.warning(id, 'Relation "{}" has no missing values '
                                           '(mask)'.format(relation_str(relation)))
                    rmses = []
                    for fuser in fusers.values():
                        rep_rmse = []
                        for fuserfit in fuser:
                            if not fuserfit.can_complete(relation):
                                break
                            completion = fuserfit.complete(relation)
                            rep_rmse.append(RMSE(relation.data, completion))
                        rmses.append(np.mean(rep_rmse) if rep_rmse else None)
                    try: min_rmse = min(e for e in rmses if e is not None)
                    except ValueError: continue # No fuser could complete this relation
                    for col, rmse in enumerate(rmses):
                        if rmse is None: continue
                        item = QtGui.QTableWidgetItem('{:.05f}'.format(rmse))
                        item.setFlags(QtCore.Qt.ItemIsEnabled)
                        if rmse == min_rmse and len(rmses) > 1:
                            item.setFont(BOLD_FONT)
                        self.setItem(row, col, item)
                self.setVerticalHeaderLabels([relation_str(i) for i in relations.values()])
                self.resizeColumnsToContents()
                self.resizeRowsToContents()

        self.table = HereTableWidget(box)

    def update(self):
        self.table.update_table(self.fusers, self.relations)

    def on_fuser_change(self, fuser, id):
        if fuser:
            N_RUNS = 1
            self.fusers[id] = [fuser.retrain() for _ in range(N_RUNS)]
        else: del self.fusers[id]
        self.update()

    def on_relation_change(self, relation, id):
        if relation: self.relations[id] = relation.relation
        else: del self.relations[id]
        self.warning(id, '')
        self.update()


def main():
    from sklearn.datasets import make_blobs
    import numpy as np
    from orangecontrib.datafusion.models import FittedFusionGraph
    from orangecontrib.datafusion.widgets.owmeanfuser import MeanFuser
    X, y = make_blobs(100, 3, centers=2, center_box=(-100, 100), cluster_std=10)
    X = X.astype(int)
    X += abs(X.min())

    nrows, ncols, _ = X.max(0)
    R1 = np.zeros((nrows + 1, ncols + 1))
    R1[X[:, 0], X[:, 1]] = X[:, 2]
    R1 = np.ma.array((R1 - R1.min()) / (R1.max() - R1.min()))

    _, ncols, nrows = X.max(0)
    R2 = np.zeros((nrows + 1, ncols + 1))
    R2[X[:, 2], X[:, 1]] = X[:, 0]
    R2 = np.ma.array((R2 - R2.min()) / (R2.max() - R2.min()))

    t1 = fusion.ObjectType('Users', 10)
    t2 = fusion.ObjectType('Movies', 30)
    t3 = fusion.ObjectType('Actors', 40)
    relations = [
        fusion.Relation(R1, t1, t2, name='like'),
        fusion.Relation(R2, t3, t2, name='feature in'),
    ]
    G = fusion.FusionGraph()
    for relation in relations:
        relation.data.mask = np.random.rand(*relation.data.shape) > .8
        G.add_relation(relation)
    fuserF = fusion.Dfmf()
    fuserF.fuse(G)

    from copy import deepcopy
    G = deepcopy(G)
    fuserC = fusion.Dfmc()
    fuserC.name = 'My dfmc<3'
    fuserC.fuse(G)

    app = QtGui.QApplication([])
    w = OWCompletionScoring()
    w.on_fuser_change(FittedFusionGraph(fuserF), fuserF.__class__.__name__)
    w.on_fuser_change(FittedFusionGraph(fuserC), fuserC.__class__.__name__)
    w.on_fuser_change(MeanFuser(0), 'meanfuser0')
    w.on_fuser_change(MeanFuser(1), 'meanfuser1')
    w.on_fuser_change(MeanFuser(2), 'meanfuser2')
    for i, relation in enumerate(relations, 1):
        w.on_relation_change(Relation(relation), i)
    w.show()
    app.exec()


if __name__ == "__main__":
    main()



import os
import csv
import random
import numpy as np
from skfusion import fusion


class ObjectType:
    Actors = fusion.ObjectType('Actors')
    Movies = fusion.ObjectType('Movies')
    Users = fusion.ObjectType('Users')
    Genres = fusion.ObjectType('Genres')


def actor_genre_matrix(actors):
    movies_genres, genres = movie_concept_matrix(input_movies=None, concept="genre")
    movies_actors, actors = movie_concept_matrix(input_movies=None, concept="actor", actors=actors)

    actors_genres = np.zeros((movies_actors.shape[1], movies_genres.shape[1]))
    for i in range(movies_actors.shape[0]):
        for j in np.nonzero(movies_actors[i]):
            actors_genres[j, :] += movies_genres[i, :]
    return actors_genres, genres


def movie_concept_matrix(input_movies, concept, actors=None):
    if input_movies is None:
        input_movies = get_all_movie_names()

    if concept == "genre":
        filename = "movies.csv"
    elif concept == "actor":
        filename = "actors.csv"
    else:
        raise ValueError("Wrong concept, must be genre or actor")
    filename = get_valid_file_path(filename)

    with open(filename, 'r', encoding="utf8") as f:
        items, all_movies = csv.reader(f, delimiter=','), set(input_movies)
        next(items)
        concepts = {line[1]: line[2].split('|') for line in items if line[1] in all_movies}

    if actors is not None:
        if isinstance(actors, int):
            all_actors = np.array(list(set(x for y in concepts.values() for x in y)))
            idx = set(random.sample(range(len(all_actors)), int(len(all_actors) * actors / 100.0)))
            actors = all_actors[sorted(idx)]

        actors = set(actors)
        for key, val in concepts.items():
            concepts[key] = list(set(val) & actors)

    all_concepts = sorted(list(set(x for y in concepts.values() for x in y)))
    idx = {name: index for name, index in zip(all_concepts, range(len(all_concepts)))}

    mov_gen = np.zeros((len(input_movies), len(all_concepts)))
    for i in range(len(input_movies)):
        for genre in concepts[input_movies[i]]:
            mov_gen[i, idx[genre]] = 1

    return mov_gen, all_concepts


def actor_matrix(mat):
    actors = np.zeros((mat.shape[1], mat.shape[1]))
    for row in mat:
        for i in np.nonzero(row)[0]:
            for j in np.nonzero(row)[0]:
                actors[i, j] += 1 if i != j else 0
    return actors


def movie_user_matrix(percentage=None, start_year=None, end_year=None):
    filename = get_valid_file_path("ratings.csv")
    users, movies, ratings, timestamps = np.loadtxt(filename, skiprows=1, delimiter=",", unpack=True)
    unique_users, unique_movies = np.unique(users.astype(int)), np.unique(movies.astype(int))

    if percentage is not None:
        idx = set(random.sample(range(len(unique_movies)), int(len(unique_movies) * percentage / 100.0)))
        filtered_movies = unique_movies[sorted(idx)]
    elif start_year is not None and end_year is not None and start_year <= end_year:
        years = get_all_movie_years()
        filtered_movies = [movie for movie, year in zip(unique_movies, years) if start_year <= year <= end_year]
    else:
        raise ValueError

    movie_idx = {movieId: index for index, movieId in enumerate(filtered_movies)}
    matrix = np.tile(np.nan, (len(filtered_movies), len(unique_users)))

    for i in range(len(ratings)):
        if movies[i] in movie_idx:
            matrix[movie_idx[movies[i]], users[i] - 1] = ratings[i]

    matrix = np.ma.masked_invalid(matrix, copy=False)
    return matrix, names_of_movies(map(str, filtered_movies)), list(map(str, list(unique_users)))


def get_all_movies():
    filename = get_valid_file_path("movies.csv")
    with open(filename, 'r', encoding="utf8") as f:
        items = csv.reader(f, delimiter=',')
        next(items)
        ids = [line[0] for line in items]
    return sorted(set(ids))


def get_all_movie_names():
    filename = get_valid_file_path("movies.csv")
    with open(filename, 'r', encoding="utf8") as f:
        items = csv.reader(f, delimiter=',')
        next(items)
        names = [line[1] for line in items]
    return names


def names_of_movies(movies):
    filename = get_valid_file_path("movies.csv")
    with open(filename, 'r', encoding="utf8") as f:
        items = csv.reader(f, delimiter=',')
        next(items)
        names = {line[0]: line[1] for line in items}
    return [names[movie] for movie in movies]


def get_all_movie_years():
    movie_names = get_all_movie_names()
    return [int(x[-5:-1]) if x[-5:-1].isdigit() else 0 for x in movie_names]


def get_valid_file_path(filename):
    return os.path.join(os.path.dirname(__file__), 'datasets', filename)


if __name__ == '__main__':
    pass




from collections import defaultdict

from PyQt4 import QtGui, QtCore
from Orange.widgets import widget, gui, settings

from skfusion import fusion
from orangecontrib.datafusion.models import Relation, FusionGraph, RelationCompleter
from orangecontrib.datafusion.widgets.owfusiongraph import bold_item, rel_shape, rel_cols

import numpy as np


class Output:
    FUSER = 'Mean-fitted fusion graph'
    RELATION = 'Relation'


class MeanBy:
    ROWS = 'Rows'
    COLUMNS = 'Columns'
    VALUES = 'All values'
    all = (COLUMNS, ROWS, VALUES)


class MeanFuser(RelationCompleter):
    def __init__(self, mean_by):
        self.axis = {
            MeanBy.ROWS: 1,
            MeanBy.COLUMNS: 0,
            MeanBy.VALUES: None}[MeanBy.all[mean_by]]
        self.mean_by = mean_by

    @property
    def name(self):
        return 'Mean by ' + MeanBy.all[self.mean_by].lower()

    def __getattr__(self, attr):
        return self

    def retrain(self):
        """Mean is deterministic, return the same Completer."""
        return self

    def can_complete(self, relation):
        """MeanFuser can complete any relation."""
        return True

    def complete(self, relation):
        """Mock ``skfusion.fusion.FusionFit.complete()``"""
        assert isinstance(relation, fusion.Relation)
        A = relation.data.copy()
        if not np.ma.is_masked(A):
            return A
        mean_value = np.nanmean(A, axis=None)
        if self.axis is None:
            # Replace the mask with mean of the matrix
            A[A.mask] = mean_value
        else:
            # Replace the mask with mean by axes
            mean = np.nanmean(A, axis=self.axis)
            # Replace any NaNs in mean with mean of the matrix
            mean[np.isnan(mean)] = mean_value
            A[A.mask] = np.take(mean, A.mask.nonzero()[not self.axis])
        return A


class OWMeanFuser(widget.OWWidget):
    name = 'Mean Fuser'
    priority = 55000
    icon = 'icons/MeanFuser.svg'
    inputs = [
        ('Fusion graph', FusionGraph, 'on_fusion_graph_change'),
        ('Relation', Relation, 'on_relation_change', widget.Multiple),
    ]
    outputs = [
        (Output.FUSER, MeanFuser, widget.Default),
        (Output.RELATION, Relation)
    ]

    want_main_area = False

    mean_by = settings.Setting(0)
    selected_relation = settings.Setting(0)

    def __init__(self):
        super().__init__()
        self.relations = defaultdict(int)
        self.id_relations = {}
        self._create_layout()
        self.commit()

    def _create_layout(self):
        self.controlArea.layout().addWidget(
            gui.comboBox(self.controlArea, self, 'mean_by',
                         box='Mean fuser',
                         label='Calculate masked values as mean by:',
                         items=MeanBy.all, callback=self.commit))
        box = gui.widgetBox(self.controlArea, 'Output completed relation')
        self.table = gui.TableWidget(box, select_rows=True)
        self.table.selectionChanged = lambda _, __: self.commit()
        self.table.setColumnFilter(bold_item, (1, 3))
        self.controlArea.layout().addStretch(1)

    def commit(self, item=None):
        self.fuser = MeanFuser(self.mean_by)
        self.send(Output.FUSER, self.fuser)
        selection = self.table.selectedRanges()
        if self.table.rowCount() and selection:
            relation = self.table.rowData(selection[0].topRow())
            data = Relation.create(self.fuser.complete(relation),
                                   relation.row_type,
                                   relation.col_type,
                                   self.graph)
        else:
            data = None
        self.send(Output.RELATION, data)

    def update_table(self):
        self.table.clear()
        for rel in self.relations:
            self.table.addRow([rel_shape(rel.data)]
                              + rel_cols(rel)
                              + ['(not masked)' if not np.ma.is_masked(rel.data) else ''],
                              data=rel)

    def _add_relation(self, relation):
        self.relations[relation] += 1

    def _remove_relation(self, relation):
        self.relations[relation] -= 1
        if not self.relations[relation]:
            del self.relations[relation]

    def on_fusion_graph_change(self, graph):
        if graph:
            self.graph = graph
            for rel in graph.relations:
                self._add_relation(rel)
        else:
            for rel in self.graph.relations:
                self._remove_relation(rel)
        self.update_table()
        self.commit()

    def on_relation_change(self, relation, id):
        try: self._remove_relation(self.id_relations.pop(id))
        except KeyError: pass
        if relation:
            self.id_relations[id] = relation.relation
            self._add_relation(relation.relation)
        self.update_table()
        self.commit()


def main():
    t1 = fusion.ObjectType('Users', 10)
    t2 = fusion.ObjectType('Movies', 30)
    t3 = fusion.ObjectType('Actors', 40)

    # test that MeanFuser completes correctly
    R = np.ma.array([[1, 1, 0],
                     [3, 0, 0]], mask=[[0, 0, 1],
                                       [0, 1, 1]], dtype=float)
    rel = fusion.Relation(R, t1, t2)
    assert (MeanFuser(0).complete(rel) == [[1, 1, 5/3],
                                           [3, 1, 5/3]]).all()
    assert (MeanFuser(1).complete(rel) == [[1, 1, 1],
                                           [3, 3, 3]]).all()
    assert (MeanFuser(2).complete(rel) == [[1,   1, 5/3],
                                           [3, 5/3, 5/3]]).all()

    R1 = np.ma.array(np.random.random((20, 20)))
    R2 = np.ma.array(np.random.random((40, 40)),
                     mask=np.random.random((40,40)) > .8)
    relations = [
        fusion.Relation(R1, t1, t2, name='like'),
        fusion.Relation(R2, t3, t2, name='feature in'),
    ]
    G = fusion.FusionGraph()
    G.add_relations_from(relations)
    app = QtGui.QApplication([])
    w = OWMeanFuser()
    w.on_fusion_graph_change(G)
    w.show()
    app.exec()


if __name__ == "__main__":
    main()



import re
import functools
from copy import copy
from itertools import count

import numpy as np

from Orange.data import Table, Domain, ContinuousVariable, StringVariable, Variable
from skfusion import fusion


GENERATE_OTYPE = (fusion.ObjectType('LatentSpace' + str(i)) for i in count())


class Relation(Table):
    """Wrapper for `skfusion.fusion.Relation`
    """

    def __new__(cls, *args, **kwargs):
        """Bypass Table.__new__."""
        return object.__new__(Relation)

    def __init__(self, relation):
        """Create a wrapper for fusion.Relation.

        Parameters:
        -----------
        relation: An instance of `skfusion.fusion.Relation`
        """
        self.relation = relation
        meta_vars, self.metas = self._create_metas(relation)
        self._Y = self.W = np.zeros((len(relation.data), 0))

        if relation.col_names is not None:
            attr_names = relation.col_names
        else:
            attr_names = range(relation.data.shape[1])
        self.domain = Domain([ContinuousVariable(name)
                              for name in map(str, attr_names)],
                             metas=meta_vars)
        Table._init_ids(self)

    @staticmethod
    def _create_metas(relation):
        metas = []
        metas_data = [[] for x in relation.data]
        if relation.row_metadata is not None:
            metadata_names = set()
            for md in relation.row_metadata:
                metadata_names.update(md.keys())
            metadata_names = sorted(metadata_names, key=str)

            metas.extend(metadata_names)
            for md, v in zip(metas_data, relation.row_metadata):
                for k in metadata_names:
                    md.append(v.get(k, np.nan))
        elif relation.row_names is not None:
            metas = [relation.row_type.name]
            metas_data = [[name] for name in relation.row_names]

        def create_var(x):
            if isinstance(x, Variable):
                return x
            else:
                return StringVariable(str(x))

        metas_vars = [create_var(x) for x in metas]
        metas = np.array(metas_data, dtype='object')
        return metas_vars, metas

    @property
    def col_type(self):
        """ Column object type"""
        return self.relation.col_type

    @property
    def row_type(self):
        """Row object type"""
        return self.relation.row_type

    @property
    def name(self):
        """Relation name"""
        return self.relation.name

    @property
    @functools.lru_cache()
    def X(self):
        """
        :return: relation data
        """
        data = self.relation.data
        if np.ma.is_masked(data):
            mask = data.mask
            data = data.data.copy()
            data[mask] = np.nan
        else:
            data = data.copy()
        return data

    def __len__(self):
        """Number of rows in relation"""
        return len(self.relation.data)

    @classmethod
    def from_table(cls, domain, source, row_indices=...):
        return Table.from_table(domain, source, row_indices)

    @classmethod
    def create(cls, data, row_type, col_type, graph=None):
        row_names = row_metadata = col_names = col_metadata = None
        if row_type:
            if graph:
                row_names = graph.get_names(row_type)
                row_metadata = graph.get_metadata(row_type)
                if not any(row_metadata):
                    row_metadata = None
        else:
            row_type = next(GENERATE_OTYPE)
        if col_type:
            if graph:
                col_names = graph.get_names(col_type)
                col_metadata = graph.get_metadata(row_type)
                if not any(col_metadata):
                    col_metadata = None
        else:
            col_type = next(GENERATE_OTYPE), None
        return Relation(fusion.Relation(data, row_type, col_type, row_names=row_names, row_metadata=row_metadata,
                                        col_names=col_names, col_metadata=col_metadata))


class RelationCompleter:
    @property
    def name(self):
        """"Completer name"""
        raise NotImplementedError()

    def retrain(self):
        """Retrain the completer using different random seed.

        Returns a new instance of the Completer.
        """
        raise NotImplementedError()

    def can_complete(self, relation):
        """Return True, if the completer had sufficient data to complete the
        given relation.
        """
        raise NotImplementedError()

    def complete(self, relation):
        """Return a completed relation.

         All masked values from the parameter relation should be replaced with the predicted ones.
         """
        raise NotImplementedError()


class FusionGraph:
    def __init__(self, fusion_graph):
        """Wrapper for skfusion FusionGraph

        :type fusion_graph: skfusion.fusion.FusionGraph
        """
        super().__init__()
        self._fusion_graph = fusion_graph

    def __getattr__(self, attr):
        return getattr(self._fusion_graph, attr)

    def get_selected_nodes(self, element_id):
        """ Return ObjectTypes from that correspond to selected `element_id`
            (in the SVG).
        """
        assert element_id.startswith('edge ') or element_id.startswith('node ')
        selected_is_edge = element_id.startswith('edge ')
        # Assumes SVG element's id attributes specify nodes `-delimited
        node_names = re.findall('`([^`]+)`', element_id)
        nodes = [self._fusion_graph.get_object_type(name) for name in node_names]
        assert len(nodes) == 2 if selected_is_edge else len(nodes) == 1
        return nodes


class FittedFusionGraph(FusionGraph, RelationCompleter):
    def __init__(self, fusion_fit):
        """Wrapper for skfusion Fusion Fit

        :type fusion_fit: skfusion.fusion.FusionFit
        """
        super().__init__(fusion_fit.fusion_graph)
        self._fusion_fit = fusion_fit

    @property
    def name(self):
        return (getattr(self._fusion_fit, 'name', '') or
                '{cls}(factors={fac}, iter={mi}, init={it})'.format(
                    cls=self._fusion_fit.__class__.__name__,
                    fac=len(self._fusion_fit.factors_),
                    mi=self._fusion_fit.max_iter,
                    it=self._fusion_fit.init_type))

    @property
    def backbones_(self):
        return self._fusion_fit.backbones_

    @property
    def factors_(self):
        return self._fusion_fit.factors_

    def backbone(self, relation):
        return self._fusion_fit.backbone(relation)

    def factor(self, object_type):
        return self._fusion_fit.factor(object_type)

    def compute_chain(self, chain, end_in_input_space=True):
        row_type = chain[0].row_type
        result = self._fusion_fit.factor(row_type)
        for rel in chain:
            result = np.dot(result, self._fusion_fit.backbone(rel))
        col_type = None
        if end_in_input_space:
            col_type = chain[-1].col_type
            result = np.dot(result, self._fusion_fit.factor(col_type).T)
        return Relation.create(result, row_type, col_type, self)

    # Relation Completer members
    def can_complete(self, relation):
        try:
            for fuser_relation in self.get_relations(relation.row_type,
                                                     relation.col_type):
                if fuser_relation._id == relation._id:
                    return True
        except fusion.DataFusionError:
            pass  # relation.{row,col}_type not in fusion graph
        return False

    def complete(self, relation):
        return self._fusion_fit.complete(relation)

    def retrain(self):
        fg = copy(self._fusion_fit)
        fg.fuse(self._fusion_graph)
        return FittedFusionGraph(fg)



import sys
from PyQt4 import QtGui
from orangecontrib.datafusion.models import Relation
from Orange.widgets.widget import OWWidget
from Orange.widgets import widget, gui, settings
from orangecontrib.datafusion import movielens

from skfusion import fusion

MOVIE_ACTORS = "Movie Actors"
ACTORS_ACTORS = "Costarring Actors"


class OWIMDbActors(OWWidget):
    name = "IMDb Actors"
    description = "Construct a movie-actor and actor-actor relation matrix."
    priority = 80000
    icon = "icons/IMDbActors.svg"
    want_main_area = False
    inputs = [("Filter", Relation, "set_data")]
    outputs = [(MOVIE_ACTORS, Relation),
               (ACTORS_ACTORS, Relation)]

    percent = settings.Setting(10)

    def __init__(self):
        super().__init__()
        self.movies = None
        self.infobox = gui.widgetBox(self.controlArea, "Select Actors")

        percent = gui.hSlider(
            gui.indentedBox(self.infobox), self, "percent",
            minValue=1, maxValue=100, step=1, ticks=10, labelFormat="%d %%")

        gui.button(self.controlArea, self, "&Apply",
                   callback=self.send_output, default=True)

        self.setSizePolicy(QtGui.QSizePolicy(QtGui.QSizePolicy.Fixed,
                                             QtGui.QSizePolicy.Fixed))

        self.setMinimumWidth(250)
        self.setMaximumWidth(250)

        self.movies = None

    def set_data(self, relation):
        if relation is not None:
            assert isinstance(relation, Relation)
            if relation.col_type == movielens.ObjectType.Movies:
                self.movies = relation.relation.col_names
            elif relation.row_type == movielens.ObjectType.Movies:
                self.movies = relation.relation.row_names
            else:
                self.error(1, "Only relations with ObjectType Movies can be used to filter actors.")

            self.send_output()

    def send_output(self):
        if self.movies is not None:
            movie_actor_mat, actors = movielens.movie_concept_matrix(self.movies, concept="actor",
                                                                     actors=self.percent)
            actor_actor_mat = movielens.actor_matrix(movie_actor_mat)

            movies_actors = fusion.Relation(movie_actor_mat.T, name='play in',
                                            row_type=movielens.ObjectType.Actors, row_names=actors,
                                            col_type=movielens.ObjectType.Movies, col_names=self.movies)
            self.send(MOVIE_ACTORS, Relation(movies_actors))

            actors_actors = fusion.Relation(actor_actor_mat, name='costar with',
                                            row_type=movielens.ObjectType.Actors, row_names=actors,
                                            col_type=movielens.ObjectType.Actors, col_names=actors)
            self.send(ACTORS_ACTORS, Relation(actors_actors))


if __name__ == "__main__":
    app = QtGui.QApplication(sys.argv)
    ow = OWIMDbActors()
    # ow.set_data(movies_users)
    ow.show()
    app.exec_()



from PyQt4.QtGui import QTableView, QGridLayout, QWidget
from PyQt4.QtCore import Qt, QSize
from Orange.data import Table, Domain
from Orange.widgets.settings import Setting, ContextSetting, PerfectDomainContextHandler
from Orange.widgets.utils.itemmodels import TableModel
from Orange.widgets.widget import OWWidget
from Orange.widgets import gui
from orangecontrib.datafusion.models import Relation
from skfusion import fusion


class Output:
    RELATION = 'Relation'


class OWTableToRelation(OWWidget):
    name = "Table to Relation"
    description = "Convert data table to relation matrix. Label matrix axis."
    priority = 50000
    icon = "icons/TableToRelation.svg"

    inputs = [("Data", Table, "set_data")]
    outputs = [(Output.RELATION, Relation)]

    settingsHandler = PerfectDomainContextHandler()

    data = None

    relation_name = ContextSetting("")
    transpose = ContextSetting(False)

    row_type = ContextSetting("")
    selected_meta = ContextSetting(0)
    row_names = None

    col_type = ContextSetting("")
    col_names = None

    auto_commit = Setting(True)

    def __init__(self):
        super().__init__()

        self.model = None
        self.view = None
        self.row_names_combo = None
        self.icons = gui.attributeIconDict
        self.populate_control_area()
        self.populate_main_area()

    def populate_control_area(self):
        rel = gui.widgetBox(self.controlArea, "Relation")
        gui.lineEdit(rel, self, "relation_name", "Name", callbackOnType=True, callback=self.apply)
        gui.checkBox(rel, self, "transpose", "Transpose", callback=self.apply)

        col = gui.widgetBox(self.controlArea, "Column")
        gui.lineEdit(col, self, "col_type", "Object Type", callbackOnType=True, callback=self.apply)

        row = gui.widgetBox(self.controlArea, "Row")
        gui.lineEdit(row, self, "row_type", "Object Type", callbackOnType=True, callback=self.apply)
        self.row_names_combo = gui.comboBox(row, self, "selected_meta", label="Object Names",
                                            callback=self.update_row_names)

        gui.rubber(self.controlArea)
        gui.auto_commit(self.controlArea, self, "auto_commit", "Send",
                        checkbox_label='Auto-send',
                        orientation='vertical')

    def populate_main_area(self):
        grid = QWidget()
        grid.setLayout(QGridLayout(grid))
        self.mainArea.layout().addWidget(grid)

        col_type = gui.label(None, self, '%(col_type)s')

        grid.layout().addWidget(col_type, 0, 1)
        grid.layout().setAlignment(col_type, Qt.AlignHCenter)

        row_type = gui.label(None, self, '%(row_type)s')
        grid.layout().addWidget(row_type, 1, 0)
        grid.layout().setAlignment(row_type, Qt.AlignVCenter)

        self.view = QTableView()
        self.model = None
        grid.layout().addWidget(self.view, 1, 1)

    def sizeHint(self):
        return QSize(800, 500)

    def set_data(self, data):
        self.closeContext()
        self.data = data
        if data is not None:
            self.init_attr_values(data.domain.metas)
            self.openContext(self.data)
            self.col_names = [str(a.name) for a in data.domain.attributes]
            if hasattr(data, 'col_type'):
                self.col_type = data.col_type
        else:
            self.init_attr_values(())
        self.update_preview()
        self.update_row_names()
        self.unconditional_commit()

    def init_attr_values(self, candidates):
        self.col_type = ""
        self.col_names = None

        if candidates:
            self.row_type = candidates[0].name
            self.selected_meta = 1
        else:
            self.row_type = ""
            self.selected_meta = 0
            self.row_names = None

        self.row_names_combo.clear()
        self.row_names_combo.addItem('(None)')
        for var in candidates:
            self.row_names_combo.addItem(self.icons[var], var.name)
        self.row_names_combo.setCurrentIndex(self.selected_meta)

    def update_row_names(self):
        if self.selected_meta:
            self.row_names = list(self.data[:, -self.selected_meta].metas.flatten())
        else:
            self.row_names = None

        if self.model:
            self.model.headerDataChanged.emit(
                Qt.Vertical, 0, self.model.rowCount() - 1)
        self.commit()

    def update_preview(self):
        this = self

        class MyTableModel(TableModel):
            def headerData(self, section, orientation, role):
                if orientation == Qt.Vertical and role == Qt.DisplayRole:
                    if this.row_names:
                        return this.row_names[section]
                else:
                    return super().headerData(section, orientation, role)

        if self.data:
            domain = Domain(self.data.domain.attributes)
            preview_data = Table(domain, self.data)
            self.model = MyTableModel(preview_data)
        else:
            self.model = None
        self.view.setModel(self.model)

    def apply(self):
        self.commit()

    def commit(self):
        if self.data:
            domain = self.data.domain
            metadata_cols = list(domain.class_vars) + list(domain.metas)
            metadata = [{var: var.to_val(value) for var, value in zip(metadata_cols, values.list)}
                        for values in self.data[:, metadata_cols]]

            if self.transpose:
                relation = fusion.Relation(
                    self.data.X.T, name=self.relation_name,
                    row_type=fusion.ObjectType(self.col_type or 'Unknown'), row_names=self.col_names,
                    col_type=fusion.ObjectType(self.row_type or 'Unknown'), col_names=self.row_names,
                    col_metadata=metadata)
            else:
                relation = fusion.Relation(
                    self.data.X, name=self.relation_name,
                    row_type=fusion.ObjectType(self.row_type or 'Unknown'), row_names=self.row_names,
                    row_metadata=metadata,
                    col_type=fusion.ObjectType(self.col_type or 'Unknown'), col_names=self.col_names,
                )
            self.send(Output.RELATION, Relation(relation))


if __name__ == '__main__':
    from PyQt4.QtGui import QApplication
    import Orange
    app = QApplication([])
    ow = OWTableToRelation()
    ow.set_data(Orange.data.Table('zoo'))
    ow.show()
    app.exec()
    ow.saveSettings()



import numpy as np

from PyQt4 import QtCore, QtGui
from Orange.widgets import widget, gui, settings

from skfusion import fusion
from orangecontrib.datafusion.models import Relation, FittedFusionGraph
from orangecontrib.datafusion.widgets import owlatentfactors
from orangecontrib.datafusion.widgets.owfusiongraph import rel_cols, bold_item


class Output:
    RELATION = 'Relation'


class OWChaining(owlatentfactors.OWLatentFactors):
    name = "Chaining"
    description = "Profile objects of one type in the feature space of " \
                  "another type through chaining of latent factors."
    priority = 30000
    icon = "icons/LatentChaining.svg"
    inputs = [("Fitted fusion graph", FittedFusionGraph, "on_fuser_change")]
    outputs = [(Output.RELATION, Relation)]

    pref_complete = settings.Setting(0)  # Complete chaining to feature space

    def __init__(self):
        super().__init__()
        self.in_selection_mode = False

    def _create_layout(self):
        self.controlArea = self
        box = gui.widgetBox(self.controlArea, margin=7)
        box.layout().addWidget(self.webview)
        box = gui.widgetBox(self.controlArea, 'Latent chains')
        self.table = gui.TableWidget(box, select_rows=True)
        self.table.setColumnFilter(bold_item, lambda col: col % 2)

        def selectionChanged(selected, _):
            if not selected:
                data = None
            else:
                chain = self.table.rowData(selected[0].top())
                self.webview.evalJS('dehighlight(ELEMENTS);')
                self._highlight_relations(chain)
                data = self.fuser.compute_chain(chain, self.pref_complete)
            self.send(Output.RELATION, data)

        self.table.selectionChanged = selectionChanged
        self.controlArea.layout().addWidget(box)

        def on_change_pref_complete():
            ranges = self.table.selectedRanges()
            self._populate_table(self.chains)
            # Re-apply selection
            if ranges:
                self.table.selectRow(ranges[0].topRow())

        gui.radioButtons(box, self, 'pref_complete',
                         label='Complete chain to:',
                         btnLabels=('latent space', 'feature space'),
                         callback=on_change_pref_complete)
        self.controlArea.layout().addStretch(1)

    def _highlight_relations(self, relations):
        selectors = set()
        for rel in relations:
            selectors.add('.node[id*={}]'.format(rel.row_type.name))
            selectors.add('.node[id*={}]'.format(rel.col_type.name))
            selectors.add('.edge[id*={}][id*={}]'.format(rel.row_type.name, rel.col_type.name))
        self.webview.evalJS('highlight("{}");'.format(','.join(selectors)))

    def _populate_table(self, chains=[]):
        self.table.clear()
        self.send(Output.RELATION, None)
        for chain in chains:
            columns = [str(self.selected_start)]
            for rel in chain:
                columns += rel_cols(rel)[1:]
            assert columns[-1] == str(self.selected_end)
            shape = (chain[0]. data.shape[0],
                     chain[-1].data.shape[1] if self.pref_complete else chain[-1].col_type.rank)
            self.table.addRow(['{}×{}'.format(*shape)] + columns, data=chain)
            self._highlight_relations(chain)

    def on_fuser_change(self, fuser):
        self.fuser = fuser
        self._populate_table()
        self.repaint()

    def on_graph_element_selected(self, element_id):
        if not element_id:
            self.in_selection_mode = False
            return self._populate_table()
        nodes = self.fuser.get_selected_nodes(element_id)
        selected_is_edge = len(nodes) > 1
        if selected_is_edge:
            self.webview.evalJS('dehighlight(ELEMENTS);')
            self.in_selection_mode = False
            return self._populate_table()
        if not self.in_selection_mode:
            self.selected_start, self.selected_end = nodes[0], None
            self.in_selection_mode = True
        else:
            self.selected_end = nodes[0]
            self.in_selection_mode = False
        if not (self.selected_start and self.selected_end):
            return

        def _get_chains(ot1, ot2):
            """ Return all chains of relations that lead from ObjectType `ot1`
                to `ot2`.
            """
            G = self.fuser
            results, paths = [], [(ot1, [])]
            while paths:
                cur, path = paths.pop()
                if cur == ot2 and path:
                    results.append(path)
                    continue
                for rel in G.out_relations(cur):
                    # Discount relations to self, constraints (= prevent cycles)
                    if rel.row_type == rel.col_type: continue
                    # Discount types that are already in path (=prevent cycles)
                    if any(rel.col_type in r for r in path): continue
                    paths.append((rel.col_type, path + [rel]))
            results.sort(key=len)
            return results
        chains = _get_chains(self.selected_start, self.selected_end)

        # Populate the listview
        self.chains = chains
        self._populate_table(chains)

        # If no chains lead from start to end, reinterpret end as start
        if not chains:
            self.on_graph_element_selected(element_id)


def main():
    # example from https://github.com/marinkaz/scikit-fusion
    import numpy as np
    R12 = np.random.rand(50, 100)
    R32 = np.random.rand(100, 150)
    R33 = np.random.rand(150, 150)
    R13 = np.random.rand(50, 150)
    t1 = fusion.ObjectType('Users', 10)
    t2 = fusion.ObjectType('Movies', 30)
    t3 = fusion.ObjectType('Actors', 40)
    relations = [fusion.Relation(R12, t1, t2, name='like'),
                 fusion.Relation(R13, t1, t3, name='are fans of'),
                 fusion.Relation(R12, t1, t2, name='don\'t like'),
                 fusion.Relation(R33, t3, t3, name='married to'),
                 fusion.Relation(R32, t2, t3, name='feature')]
    G = fusion.FusionGraph()
    for rel in relations:
        G.add_relation(rel)
    fuser = fusion.Dfmf()
    fuser.fuse(G)
    app = QtGui.QApplication([])
    w = OWChaining()
    w.on_fuser_change(FittedFusionGraph(fuser))
    w.show()
    app.exec()


if __name__ == "__main__":
    main()



# Category metadata.

# Category icon show in the menu
ICON = "icons/Category-Fusion.svg"

# Background color for category background in menu
# and widget icon background in workflow.
BACKGROUND = "light-green"



import sys
import copy
import Orange.data.table
from PyQt4 import QtGui
from PyQt4.QtCore import Qt
from Orange.widgets.widget import OWWidget
from Orange.widgets import widget, gui, settings
from orangecontrib.datafusion.models import Relation

import numpy as np


class Output:
    IN_SAMPLE_DATA = "In-sample Data"
    OUT_OF_SAMPLE_DATA = "Out-of-sample Data"


class SampleBy:
    ROWS = 'Rows'
    COLS = 'Columns'
    ROWS_COLS = 'Rows and columns'
    ENTRIES = 'Entries'
    all = [ROWS, COLS, ROWS_COLS, ENTRIES]


def hide_data(table, percentage, sampling_type):
    assert not np.ma.is_masked(table)
    np.random.seed(0)
    if sampling_type == SampleBy.ROWS_COLS:

        row_s_mask, row_oos_mask = hide_data(table, np.sqrt(percentage), SampleBy.ROWS)
        col_s_mask, col_oos_mask = hide_data(table, np.sqrt(percentage), SampleBy.COLS)

        oos_mask = np.logical_and(row_oos_mask, col_oos_mask)
        return oos_mask

    elif sampling_type == SampleBy.ROWS:
        rand = np.repeat(np.random.rand(table.X.shape[0], 1), table.X.shape[1], axis=1)
    elif sampling_type == SampleBy.COLS:
        rand = np.repeat(np.random.rand(1, table.X.shape[1]), table.X.shape[0], axis=0)
    elif sampling_type == SampleBy.ENTRIES:
        rand = np.random.rand(*table.X.shape)
    else:
        raise ValueError("Unknown sampling method.")

    oos_mask = np.logical_and(rand >= percentage, ~np.isnan(table))
    return oos_mask


class OWSampleMatrix(OWWidget):
    name = "Matrix Sampler"
    description = "Sample a relation matrix."
    priority = 60000
    icon = "icons/MatrixSampler.svg"
    want_main_area = False
    inputs = [("Data", Orange.data.table.Table, "set_data", widget.Default)]
    outputs = [(Output.IN_SAMPLE_DATA, Relation),
               (Output.OUT_OF_SAMPLE_DATA, Relation)]

    percent = settings.Setting(90)
    method = settings.Setting(0)
    bools = settings.Setting([])

    def __init__(self):
        super().__init__()
        self.relation = None

        form = QtGui.QGridLayout()
        methodbox = gui.radioButtonsInBox(
            self.controlArea, self, "method", [],
            box=self.tr("Sampling method"), orientation=form)

        rows = gui.appendRadioButton(methodbox, "Rows", addToLayout=False)
        form.addWidget(rows, 0, 0, Qt.AlignLeft)

        cols = gui.appendRadioButton(methodbox, "Columns", addToLayout=False)
        form.addWidget(cols, 0, 1, Qt.AlignLeft)

        rows_and_cols = gui.appendRadioButton(methodbox, "Rows and columns", addToLayout=False)
        form.addWidget(rows_and_cols, 1, 0, Qt.AlignLeft)

        entries = gui.appendRadioButton(methodbox, "Entries", addToLayout=False)
        form.addWidget(entries, 1, 1, Qt.AlignLeft)

        sample_size = gui.widgetBox(self.controlArea, "Proportion of data in the sample")
        percent = gui.hSlider(
            sample_size, self, 'percent', minValue=1, maxValue=100, step=5,
            ticks=10, labelFormat=" %d%%")

        gui.button(self.controlArea, self, "&Apply",
                   callback=self.send_output, default=True)

        self.setSizePolicy(
            QtGui.QSizePolicy(QtGui.QSizePolicy.Fixed,
                              QtGui.QSizePolicy.Fixed))

        self.setMinimumWidth(250)
        self.send_output()

    def set_data(self, data):
        self.relation = data
        self.send_output()

    def send_output(self):
        if self.relation is not None:
            oos_mask = hide_data(self.relation,
                                 self.percent / 100,
                                 SampleBy.all[self.method])
            def _mask_relation(relation, mask):
                if np.ma.is_masked(relation.data):
                    mask = np.logical_or(mask, relation.data.mask)
                data = copy.copy(relation)
                data.data = np.ma.array(data.data, mask=mask)
                return data

            oos_mask = _mask_relation(self.relation.relation, oos_mask)

            self.send(Output.IN_SAMPLE_DATA, Relation(oos_mask))
            self.send(Output.OUT_OF_SAMPLE_DATA, Relation(oos_mask))


if __name__ == "__main__":
    app = QtGui.QApplication(sys.argv)
    ow = OWSampleMatrix()
    # ow.set_data(Orange.data.Table("housing.tab"))
    ow.send_output()
    ow.show()
    app.exec_()
